//////////////////////////////////////////////////////////////////////////////
//
//  Xilinx, Inc. 2002                 www.xilinx.com
//
//  XAPP 291 - Self Addressing FIFO
//
//////////////////////////////////////////////////////////////////////////////
//
//  File name :       saf_2to1.v
//
//  Description :     Top-level for 2:1 SAF
//		      addrsize used to define the fifo depth(ie number of address bits), minimum 2,
//		      maximum 26 with ram in 1024x18/512x36 mode
// 		      datasize defines the input data path width, minimum 2, maximum 26
// 		      addrsize + datasize must be less than or equal 18
// 		      addrsize + datasize must be less than or equal 30
// 		      This example uses addrsize = 9, and datasize = 16
// 		      flag(0) high when fifo between empty and one half full
// 		      flag(1) high when fifo between one quarter and three quarters full
// 		      flag(2) high when fifo between one half full and full
//
//  Date - revision : December 11th 2002 - v 1.2
//
//  Author :          NJS
//
//  Contact : e-mail  hotline@xilinx.com
//            phone   + 1 800 255 7778
//
//  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are
//              provided to you "as is". Xilinx and its licensors make and you
//              receive no warranties or conditions, express, implied,
//              statutory or otherwise, and Xilinx specifically disclaims any
//              implied warranties of merchantability, non-infringement,or
//              fitness for a particular purpose. Xilinx does not warrant that
//              the functions contained in these designs will meet your
//              requirements, or that the operation of these designs will be
//              uninterrupted or error free, or that defects in the Designs
//              will be corrected. Furthermore, Xilinx does not warrantor
//              make any representations regarding use or the results of the
//              use of the designs in terms of correctness, accuracy,
//              reliability, or otherwise.
//
//              LIMITATION OF LIABILITY. In no event will Xilinx or its
//              licensors be liable for any loss of data, lost profits,cost
//              or procurement of substitute goods or services, or for any
//              special, incidental, consequential, or indirect damages
//              arising from the use or operation of the designs or
//              accompanying documentation, however caused and on any theory
//              of liability. This limitation will apply even if Xilinx
//              has been advised of the possibility of such damage. This
//              limitation shall apply not-withstanding the failure of the
//              essential purpose of any limited remedies herein.
//
//  Copyright  2002 Xilinx, Inc.
//  All rights reserved
//
//////////////////////////////////////////////////////////////////////////////
//
`timescale 1 ps / 1ps

module saf_2to1 (
	wr_clk,					// write clock input
	rd_clk,					// read clock input
	datain,					// data input
	rst,					// reset input
	we,					// write enable input
	oe,					// output enable input
	flag,					// flag outputs
	dataout);				// output data

input 		wr_clk, rd_clk, rst, we, oe ;
input 	[15:0] 	datain;

output 	[2:0] 	flag;
output 	[7:0] 	dataout;

wire 		high ;
wire 		low ;
wire 		oe_int ;
wire 		we_int ;
wire 	[0:0] 	low1 ;
wire 	[1:0] 	low2 ;
wire 	[3:0] 	low4 ;
wire 	[7:0] 	low8 ;
wire 	[15:0] 	low16 ;
wire 	[17:0] 	douta ;
wire 	[35:0] 	doutb ;
reg 	[35:0] 	dinb ;
wire 	[15:0] 	ram_douta ;
wire 	[31:0] 	ram_doutb  ;
wire 	[1:0] 	ram_pouta ;
wire 	[3:0] 	ram_poutb ;
wire 	[1:0] 	ram_pina ;
wire 	[3:0] 	ram_pinb ;
wire 	[31:0] 	ram_dinb ;
reg 	[9:0] 	ram_addra ;
reg 	[8:0] 	ram_addrb  ;
wire 	[9:0] 	next_wr_addressa ;
wire 	[9:0] 	next_wr_addressb ;
wire 	[9:0] 	tempa ;

assign low = 1'b0 ;
assign low1 = 1'b1 ;
assign low2 = 2'b00 ;
assign low4 = 4'b0000 ;
assign low8 = 8'h00 ;
assign low16 = 16'h0000 ;
assign high = 1'b1 ;
assign oe_int = oe | rst ; //needed to make sure the ram enable is high during reset (because its a synchronous reset)
assign we_int = we | rst ; //needed to make sure the ram enable is high during reset (because its a synchronous reset)

assign douta = {ram_pouta, ram_douta} ;
assign doutb = {ram_poutb, ram_doutb} ;

always @ (next_wr_addressa or next_wr_addressb or datain) // need to write address in two places, one ends '0' the other '1'
begin
dinb[35:34] = next_wr_addressb[9:8] ;
dinb[33:32] = next_wr_addressa[9:8] ;
dinb[31:16] = {next_wr_addressb[7:0], datain[15:8]};
dinb[15:0] = {next_wr_addressa[7:0], datain[7:0]} ;
end

assign ram_pinb = dinb[35:32] ;
assign ram_dinb = dinb[31:0] ;

always @ (douta)
begin
ram_addra[9:0] = douta[17:8] ;
end

always @ (doutb)
begin
ram_addrb[8:0] = {doutb[35:34], doutb[31:25]};
end

assign dataout = ram_douta[7:0] ;

assign flag[2] =(((ram_pouta[1:0] == 2'b00) && (ram_poutb[3:2] == 2'b11)) |
		 ((ram_pouta[1:0] == 2'b01) && (ram_poutb[3:2] == 2'b10)) |
		 ((ram_pouta[1:0] == 2'b11) && (ram_poutb[3:2] == 2'b00)) |
		 ((ram_pouta[1:0] == 2'b10) && (ram_poutb[3:2] == 2'b01)))
		 ? 1'b1 : 1'b0;

assign flag[1] =(((ram_pouta[1:0] == 2'b00) && (ram_poutb[3:2] == 2'b01)) |
		 ((ram_pouta[1:0] == 2'b01) && (ram_poutb[3:2] == 2'b11)) |
		 ((ram_pouta[1:0] == 2'b11) && (ram_poutb[3:2] == 2'b10)) |
		 ((ram_pouta[1:0] == 2'b10) && (ram_poutb[3:2] == 2'b00)))
		 ? 1'b1 : 1'b0;

assign flag[0] =(((ram_pouta[1:0] == 2'b00) && (ram_poutb[3:2] == 2'b00)) |
		 ((ram_pouta[1:0] == 2'b01) && (ram_poutb[3:2] == 2'b01)) |
		 ((ram_pouta[1:0] == 2'b11) && (ram_poutb[3:2] == 2'b11)) |
		 ((ram_pouta[1:0] == 2'b10) && (ram_poutb[3:2] == 2'b10)))
		 ? 1'b1 : 1'b0;

RAMB16_S18_S36 block_ram (
	.DIA(low16),
	.DIB(ram_dinb),
	.DIPA(low2),
	.DIPB(ram_pinb),
	.ENA(oe_int),
	.ENB(we_int),
	.WEA(low),
	.WEB(we),
	.SSRA(rst),
	.SSRB(rst),
	.CLKA(rd_clk),
	.CLKB(wr_clk),
	.ADDRA(ram_addra),
	.ADDRB(ram_addrb),
	.DOA(ram_douta),
	.DOB(ram_doutb),
	.DOPA(ram_pouta),
	.DOPB(ram_poutb));

assign tempa = {doutb[35:34], doutb[31:24]} ;

incrementer_10bit inca (
	.ip(tempa),
	.op(next_wr_addressa));

incrementer_10bit incb (
	.ip(next_wr_addressa),
	.op(next_wr_addressb));

endmodule

